@@ -1,5 +1,6 @@
 import asyncio
 from enum import Enum
+from typing import Any, Dict, Optional
 from uuid import uuid4
 
 PRIORITY_NORMAL = "5"
@@ -29,14 +30,14 @@ class NotificationRequest:
 
     def __init__(
         self,
-        device_token,
-        message,
-        notification_id=None,
-        time_to_live=None,
-        priority=None,
-        collapse_key=None,
-        push_type=None,
-    ):
+        device_token: str,
+        message: Dict[str, Any],
+        notification_id: Optional[str] = None,
+        time_to_live: Optional[int] = None,
+        priority: Optional[int] = None,
+        collapse_key: Optional[str] = None,
+        push_type: Optional[PushType] = None,
+    ) -> None:
         self.device_token = device_token
         self.message = message
         self.notification_id = notification_id or str(uuid4())
@@ -47,25 +48,34 @@ class NotificationRequest:
 
 
 class NotificationResult:
-    __slots__ = ("notification_id", "status", "description")
+    __slots__ = ("notification_id", "status", "description", "timestamp")
 
-    def __init__(self, notification_id, status, description=None):
+    def __init__(
+        self,
+        notification_id: str,
+        status: str,
+        description: Optional[str] = None,
+        timestamp: Optional[int] = None,
+    ):
         self.notification_id = notification_id
         self.status = status
         self.description = description
+        self.timestamp = timestamp
 
     @property
-    def is_successful(self):
+    def is_successful(self) -> bool:
         return self.status == APNS_RESPONSE_CODE.SUCCESS
 
 
 class DynamicBoundedSemaphore(asyncio.BoundedSemaphore):
+    _bound_value: int
+
     @property
-    def bound(self):
+    def bound(self) -> int:
         return self._bound_value
 
     @bound.setter
-    def bound(self, new_bound):
+    def bound(self, new_bound: int) -> None:
         if new_bound > self._bound_value:
             if self._value > 0:
                 self._value += new_bound - self._bound_value
@@ -76,13 +86,13 @@ class DynamicBoundedSemaphore(asyncio.BoundedSemaphore):
             self._value -= self._bound_value - new_bound
         self._bound_value = new_bound
 
-    def release(self):
+    def release(self) -> None:
         self._value += 1
         if self._value > self._bound_value:
             self._value = self._bound_value
         self._wake_up_next()
 
-    def destroy(self, exc):
+    def destroy(self, exc: Exception) -> None:
         while self._waiters:
             waiter = self._waiters.popleft()
             if not waiter.done():
